package app;

// Java imports
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

// Project imports
import util.*;

// Memory viewer and hex editor window
class Memory extends ChildWindow {

    // Private fields
    private int  address; // Address of top row
    private Font font;    // Display font

    // UI components
    private JPanel client;     // Client area
    private JLabel fontHeight; // Font height proxy
    private ArrayList<Row> rows; // Rows of text



    ///////////////////////////////////////////////////////////////////////////
    //                                Classes                                //
    ///////////////////////////////////////////////////////////////////////////

    // One row of output
    private class Row {
        JLabel   address; // Address (row header)
        JLabel[] bytes;   // Hexadecimal bytes
    }



    ///////////////////////////////////////////////////////////////////////////
    //                             Constructors                              //
    ///////////////////////////////////////////////////////////////////////////

    // Default constructor
    Memory(MainWindow parent) {
        super(parent, "memory.title");

        // Configure instance fields
        address    = 0x00000000;
        font       = new Font(Util.fontFamily(new String[]
            { "Consolas", Font.MONOSPACED } ), Font.PLAIN, 14);
        fontHeight = new JLabel(".");
        rows       = new ArrayList<Row>();

        // Configure client area
        client = new JPanel(null);
        client.addComponentListener(Util.onResize(e->onResize()));
        client.addKeyListener(Util.onKey(e->onKeyDown(e), null));
        client.addMouseWheelListener(e->onWheel(e));
        client.setFocusable(true);
        client.setPreferredSize(new Dimension(640, 480));
        client.setBackground(SystemColor.window);

        // Configure component
        setContentPane(client);
        setFont2(font);
        pack();
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Package Methods                            //
    ///////////////////////////////////////////////////////////////////////////

    // Update the display
    void refresh() {

        // The element is not ready
        if (client == null)
            return;

        // Configure working variables
        int height     = client.getHeight();
        int lineHeight = fontHeight.getPreferredSize().height;
        int count      = (height + lineHeight - 1) / lineHeight;
        var data       = new byte[count * 16];
        var widths     = new int[2];

        // Retrieve all visible bytes from the emulation context
        parent.vue.read(address, data, 0, data.length);

        // Update visible rows
        for (int x = 0; x < count; x++) {
            Row row;

            // Retrieve the row if it exists, make a new one otherwise
            if (x < rows.size())
                row = rows.get(x);
            else row = createRow();

            // Configure the row
            update(row, address + x * 16, data, x * 16);
            setVisible(row, true);
            measure(row, widths);
        }

        // Hide any rows that are not visible
        for (int x = count; x < rows.size(); x++)
            setVisible(rows.get(x), false);

        // Position components
        for (int x = 0; x < rows.size(); x++)
            arrange(rows.get(x), x * lineHeight, widths, lineHeight);

        // Finalize layout
        client.revalidate();
        client.repaint();
    }

    // Specify a new font
    void setFont2(Font font) {
        this.font = font;
        fontHeight.setFont(font);
        for (var row : rows) {
            row.address.setFont(font);
            for (var label : row.bytes)
                label.setFont(font);
        }
        onResize();
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Event Handlers                             //
    ///////////////////////////////////////////////////////////////////////////

    // Key down
    private void onKeyDown(KeyEvent e) {
        int     code = e.getKeyCode();
        int     mods = e.getModifiersEx();
        boolean alt  = (mods & InputEvent.ALT_DOWN_MASK ) != 0;
        boolean ctrl = (mods & InputEvent.CTRL_DOWN_MASK) != 0;
        int     tall = Math.max(1, tall(false));

        // No Alt combinations
        if (alt) return;

        // Goto
        if (ctrl && code == KeyEvent.VK_G) {
            String addr = JOptionPane.showInputDialog(
                this, "Goto:", "Goto", JOptionPane.PLAIN_MESSAGE);
            if (addr != null && addr.trim().length() != 0) {
                try { setAddress((int) Long.parseLong(addr, 16)); }
                catch (Exception x) { }
            }
            return;
        }

        // Seek
        switch (code) {
            case KeyEvent.VK_UP       : setAddress(address -        16); break;
            case KeyEvent.VK_DOWN     : setAddress(address +        16); break;
            case KeyEvent.VK_PAGE_UP  : setAddress(address - tall * 16); break;
            case KeyEvent.VK_PAGE_DOWN: setAddress(address + tall * 16); break;
        }

    }

    // Client resize
    private void onResize() {
        refresh();
    }

    // Mouse wheel
    private void onWheel(MouseWheelEvent e) {
        int     amount = e.getUnitsToScroll();
        int     mods   = e.getModifiersEx();
        boolean alt    = (mods & InputEvent.ALT_DOWN_MASK ) != 0;
        boolean ctrl   = (mods & InputEvent.CTRL_DOWN_MASK) != 0;

        // No Alt or Ctrl combinations
        if (amount == 0 || alt || ctrl)
            return;

        // Seek
        setAddress(address + 16 * amount);
    }



    ///////////////////////////////////////////////////////////////////////////
    //                            Private Methods                            //
    ///////////////////////////////////////////////////////////////////////////

    // Position the elements of a row
    private void arrange(Row row, int y, int[] widths, int height) {
        int spacing = (height + 1) / 2;

        // Size and position the address header
        row.address.setSize(row.address.getPreferredSize());
        row.address.setLocation(0, y);

        // Size and position the byte labels
        for (int z = 0, x = widths[0] + height; z < 16; z++) {
            var label = row.bytes[z];
            label.setBounds(x, y, widths[1], height);
            x += widths[1] + (z == 7 ? height : spacing);
        }

    }

    // Add a new row of output
    private Row createRow() {
        var row = new Row();
        row.address =

        // Address label
        row.address = new JLabel();
        row.address.setFont(font);
        row.address.setForeground(SystemColor.windowText);
        row.address.setVisible(false);
        client.add(row.address);

        // Byte labels
        row.bytes = new JLabel[16];
        for (int x = 0; x < row.bytes.length; x++) {
            var label = row.bytes[x] = new JLabel();
            label.setFont(font);
            label.setForeground(SystemColor.windowText);
            label.setVisible(false);
            client.add(label);
        }

        // Add the row to the collection
        rows.add(row);
        return row;
    }

    // Measure the minimum column widths for a row
    private void measure(Row row, int[] widths) {
        widths[0] = Math.max(widths[0], row.address.getPreferredSize().width);
        for (int x = 0; x < 16; x++)
            widths[1] = Math.max(widths[1],
                row.bytes[x].getPreferredSize().width);
    }

    // Specify the address of the top row of output
    private void setAddress(int address) {
        this.address = address & 0xFFFFFFF0;
        refresh();
    }

    // Show or hide a row
    private void setVisible(Row row, boolean visible) {
        row.address.setVisible(visible);
        for (var label : row.bytes)
            label.setVisible(visible);
    }

    // Determine how many rows of output are visible
    private int tall(boolean partial) {
        int lineHeight = fontHeight.getPreferredSize().height;
        return lineHeight == 0 ? 0 :
            (client.getHeight() + (partial ? lineHeight + 1 : 0)) / lineHeight;
    }

    // Update the text of a row
    private void update(Row row, int address, byte[] data, int offset) {
        row.address.setText(String.format("%08X", address));
        for (var label : row.bytes)
            label.setText(String.format("%02X", data[offset++] & 0xFF));
    }

}
